/**
* @file sample_process.cpp
*
* Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/
#include "sample_process.h"
#include <iostream>
#include <vector>
#include <cmath>
#include <numeric>
#include <string>
#include <algorithm>
#include <fstream>
#include <ctime>
#include <fstream>
#include <filesystem>
#include "/usr/local/mdc_sdk_llvm/dp_gea/mdc_cross_compiler/sysroot/usr/include/opencv4/opencv2/imgcodecs.hpp"
#include "/usr/local/mdc_sdk_llvm/dp_gea/mdc_cross_compiler/sysroot/usr/include/opencv4/opencv2/imgproc.hpp"
#include "preprocess.h"
#include "model_process.h"
#include "acl/acl.h"
#include "utils.h"
#include "postprocess.h"

#define CLOCK_Enable

using namespace std;
// namespace fs = std::filesystem;
namespace fs = std::__fs::filesystem;

extern bool g_isDevice;

struct rgb{
    int r;
    int g;
    int b;
};

std::string folderPath = "/home/mdc/mohsen/resnet50_11/data/CAM_BACK/";

std::string outputFolderPath = "/home/mdc/mohsen/resnet50_11/data/CAM_BACK_fp32/";

SampleProcess::SampleProcess() :deviceId_(0), context_(nullptr), stream_(nullptr)
{
}

SampleProcess::~SampleProcess()
{
    DestroyResource();
}

Result SampleProcess::InitResource()
{
    // ACL init
    // acl.json is dump or profiling config file
    if (!Utils::validStatus("aclInit", aclInit(nullptr))) {
        return FAILED;
    }

    if (!Utils::validStatus("aclrtSetDevice", aclrtSetDevice(deviceId_))) {
        ERROR_LOG("DeviceId is %d", deviceId_);
        (void)aclFinalize();
        return FAILED;
    }
    const char* soc_name{ aclrtGetSocName() };
    if (soc_name == nullptr)
    {
        ERROR_LOG("Can not get the soc name");
    }
    INFO_LOG("soc name: %s", soc_name);

    // create context (set current)
    if (!Utils::validStatus("aclrtCreateContext", aclrtCreateContext(&context_, deviceId_))) {
        ERROR_LOG("deviceId = %d", deviceId_);
        return FAILED;
    }

    // create stream
    if (!Utils::validStatus("aclrtCreateStream", aclrtCreateStream(&stream_))) {
        ERROR_LOG("deviceId = %d", deviceId_);
        return FAILED;
    }

    // get run mode
    // runMode is ACL_HOST which represents app is running in host
    // runMode is ACL_DEVICE which represents app is running in device
    aclrtRunMode runMode;
    if (!Utils::validStatus("aclrtGetRunMode", aclrtGetRunMode(&runMode))) {
        return FAILED;
    }
    g_isDevice = (runMode == ACL_DEVICE);
    return SUCCESS;
}

Result SampleProcess::Process()
{
    ModelProcess modelProcess;

    // model preprocessing
    // cv::Mat image;
    // image=cv::imread("/home/mdc/mohsen/resnet50_11/data/CAM_FRONT_LEFT/n008-2018-08-01-15-16-36-0400__CAM_FRONT_LEFT__1533151608504799.jpg");

    // model init
    
    const char* omModelPath = "/home/mdc/mohsen/resnet50_11/bin/model/fcos3d_aug_resnet18_epoch20_sim_fp32.om";
    Result ret = modelProcess.LoadModel(omModelPath);
    if (ret != SUCCESS) {
        ERROR_LOG("execute LoadModel failed");
        return FAILED;
    }

    ret = modelProcess.CreateModelDesc();
    if (ret != SUCCESS) {
        ERROR_LOG("execute CreateModelDesc failed");
        return FAILED;
    }


    size_t devBufferSize;
    void *picDevBuffer = nullptr;
    ret = modelProcess.GetInputSizeByIndex(0, devBufferSize);
    if (ret != SUCCESS) {
        ERROR_LOG("execute GetInputSizeByIndex failed");
        return FAILED;
    }

    aclError aclRet = aclrtMalloc(&picDevBuffer, devBufferSize, ACL_MEM_MALLOC_NORMAL_ONLY);
    if (!Utils::validStatus("aclrtMalloc", aclRet)) {
        ERROR_LOG("input for model size is %zu", devBufferSize);
        return FAILED;
    }

    
    vector <DataInfo> inputData = {{picDevBuffer, devBufferSize}};



    ret = modelProcess.CreateInput(inputData);
    if (ret != SUCCESS) {
        ERROR_LOG("execute CreateInput failed");
        aclrtFree(picDevBuffer);
        return FAILED;
    }




    ret = modelProcess.CreateOutput();
    if (ret != SUCCESS) {
        aclrtFree(picDevBuffer);
        ERROR_LOG("execute CreateOutput failed");
        return FAILED;
    }

    // #ifdef CLOCK_Enable
    clock_t start,end,start_preprocess,end_preprocess,start_postprocess,end_postprocess;
    double total_time_preprocess =0;
    double total_time =0;
    double total_time_postprocess =0;
    double total_time_procedure =0;
    // for(size_t i=0; i<20; i++){
    
    // #endif

    WARN_LOG("start to preprocess of model.");

    for (const auto& entry : fs::directory_iterator(folderPath)) {
    if (entry.is_regular_file() && entry.path().extension() == ".jpg") {
        std::string filePath = entry.path().string();
        cv::Mat image = cv::imread(filePath);
        if (image.empty()) {
            std::cerr << "Failed to read image: " << filePath << std::endl;
            continue; // Skip to the next image if loading fails
        }
    // cv::Mat image = cv::imread("/home/mdc/mohsen/resnet50_11/data/CAM_BACK/n008-2018-08-01-15-16-36-0400__CAM_BACK__1533151603537558.jpg");
    // std::cout << "image mat befor the preprocess: " << std::endl << image << std::endl;

    // auto VecImage = convert_cv_to_vector(image);
    std::vector<uchar>* imageVectorPtr = matToVectorPointer(image);
    

    cv::Mat imageMat = vectorPointerToMat(imageVectorPtr, 900, 1600);


    start_preprocess = clock();
    auto output_of_preprocess = preprocess_func(imageVectorPtr);
    end_preprocess = clock();
    delete imageVectorPtr;
    WARN_LOG("start to prepare inputs of model from preprocess");

    ret = Utils::MemcpyVariableToDeviceBuffer(output_of_preprocess, picDevBuffer, devBufferSize);

        if (ret != SUCCESS) {
            aclrtFree(picDevBuffer);
            ERROR_LOG("memcpy device buffer failed, index is %d", 0);
            return FAILED;
        }
        start = clock();
        // End_preprocess = clock();
        ret = modelProcess.Execute();
        end = clock();
        if (ret != SUCCESS) {
            ERROR_LOG("execute inference failed");
            aclrtFree(picDevBuffer);
            return FAILED;
        }

    auto model_output = modelProcess.OutputModelResult();
    

    start_postprocess = clock();
    auto final_results = postprocess_func(model_output);
    end_postprocess = clock();
    std::string imageName = entry.path().filename().stem(); // Get the filename without extension
        std::string outputFilePath_bboxes = outputFolderPath + imageName + "_bboxes" + ".txt";
        std::string outputFilePath_scores = outputFolderPath + imageName + "_scores" + ".txt";
        std::string outputFilePath_labels = outputFolderPath + imageName + "_labels" + ".txt";

        std::ofstream outputFile_bboxes(outputFilePath_bboxes);
        std::ofstream outputFile_scores(outputFilePath_scores);
        std::ofstream outputFile_labels(outputFilePath_labels);

        if (outputFile_bboxes.is_open()) {

            outputFile_bboxes << final_results.bboxes << std::endl;
            outputFile_bboxes.close();

        } else {
            std::cerr << "Failed to open output file: " << outputFilePath_bboxes << std::endl;
        }

        if (outputFile_scores.is_open()) {

            outputFile_scores << final_results.scores << std::endl;
            outputFile_scores.close();

        } else {
            std::cerr << "Failed to open output file: " << outputFilePath_scores << std::endl;
        }

        if (outputFile_labels.is_open()) {

            outputFile_labels << final_results.labels << std::endl;
            outputFile_labels.close();

        } else {
            std::cerr << "Failed to open output file: " << outputFilePath_labels << std::endl;
        }

    }

    #ifdef CLOCK_Enable
    
    double cpu_time_total = ((double) (end - start)) * 1000 / CLOCKS_PER_SEC;
    total_time += cpu_time_total;
    double cpu_time_total_preprocess = ((double) (end_preprocess - start_preprocess)) * 1000 / CLOCKS_PER_SEC;
    total_time_preprocess += cpu_time_total_preprocess;
    double cpu_time_total_postprocess = ((double) (end_postprocess - start_postprocess)) * 1000 / CLOCKS_PER_SEC;
    total_time_postprocess += cpu_time_total_postprocess;
    double cpu_time_total_procedure = ((double) (end_postprocess - start_preprocess)) * 1000 / CLOCKS_PER_SEC;
    total_time_procedure += cpu_time_total_procedure;
}

    // }

    std::cout << "time of preprocess: " << total_time_preprocess/81 << std::endl;
    std::cout << "time of execution: " << total_time/81 << std::endl;
    std::cout << "time of postprocess: " << total_time_postprocess/81 << std::endl;
    std::cout << "time of time of process: " << total_time_procedure/81 << std::endl;
    #endif

    modelProcess.DestroyInput();
    modelProcess.DestroyOutput();

    aclrtFree(picDevBuffer);
    return SUCCESS;
}

void SampleProcess::DestroyResource()
{
    aclError ret;
    if (stream_ != nullptr) {
        ret = aclrtDestroyStream(stream_);
        if (ret != ACL_SUCCESS) {
            ERROR_LOG("destroy stream failed, errorCode = %d", static_cast<int32_t>(ret));
        }
        stream_ = nullptr;
    }
    INFO_LOG("end to destroy stream");

    if (context_ != nullptr) {
        ret = aclrtDestroyContext(context_);
        if (ret != ACL_SUCCESS) {
            ERROR_LOG("destroy context failed, errorCode = %d", static_cast<int32_t>(ret));
        }
        context_ = nullptr;
    }
    INFO_LOG("end to destroy context");

    ret = aclrtResetDevice(deviceId_);
    if (ret != ACL_SUCCESS) {
        ERROR_LOG("reset device %d failed, errorCode = %d", deviceId_, static_cast<int32_t>(ret));
    }
    INFO_LOG("end to reset device %d", deviceId_);

    ret = aclFinalize();
    if (ret != ACL_SUCCESS) {
        ERROR_LOG("finalize acl failed, errorCode = %d", static_cast<int32_t>(ret));
    }
    INFO_LOG("end to finalize acl");
}
